#undef ARDUINO_M5STACK_FIRE
#define ARDUINO_M5STACK_Paper

#define YAML_DISABLE_CJSON

#include <ArduinoJson.h>
#include <ArduinoYaml.h>
#include <WiFi.h>
#include <M5EPD.h>
#include <NTPClient.h>
#include <HTTPClient.h>
#include <PubSubClient.h>

#include <totp.hpp>

#define LGFX_M5PAPER
#define LGFX_USE_V1
#define LGFX_AUTODETECT

#include <LovyanGFX.hpp>

static LGFX gfx;
static WiFiUDP ntpUDP;
static NTPClient timeClient(ntpUDP);

constexpr float FONT_SIZE_NORMAL = 4.0;
constexpr float FONT_SIZE_LARGE = 8.0;

static YAMLNode *SETTINGS = nullptr;

static String oldinput = "";
static String input = "";

static int last_index = -1;

static bool is_disarmed = false;

static const int WIFI_CONNECT_RETRY_MAX = 30;
static const float_t BUTTON_WIDTH = 115;
static const float_t BUTTON_HEIGHT = 115;

void drawLockState(String state)
{
  state.replace("_", " ");

  gfx.setTextSize(FONT_SIZE_NORMAL);

  gfx.setColor(TFT_WHITE);
  gfx.fillRect(405, 155, gfx.fontWidth() * 10, gfx.fontHeight());
  gfx.setColor(TFT_BLACK);

  gfx.setCursor(405, 155);
  state[0] = toupper(state[0]);
  gfx.printf("%s", state);
}

void mqtt_callback(char *topic, byte *payload, unsigned int length)
{
  String state;
  for (int i = 0; i < length; i++)
  {
    state += (const char)payload[i];
  }

  is_disarmed = state == "disarmed";

  Serial.printf("%d: %s - %s\n", length, topic, state);
  drawLockState(state);
}

WiFiClient wifiClient;
PubSubClient *client = nullptr;

void cls()
{
  gfx.setEpdMode(epd_mode_t::epd_quality);
  gfx.fillScreen(TFT_WHITE);
  gfx.waitDisplay();
  gfx.fillScreen(TFT_WHITE);
  gfx.setEpdMode(epd_mode_t::epd_fast);
  gfx.setColor(TFT_BLACK);
  gfx.waitDisplay();
}

void drawButton(int index, String label)
{
  float_t x = (125 * (index % 3)) + 15;
  float_t y = (125 * (index / 3)) + 20;

  gfx.setCursor(x + (115 / 2) - (gfx.fontWidth() / 2), y + (115 / 2) - (gfx.fontHeight() / 2));
  gfx.print(label);

  gfx.setColor(TFT_BLACK);
  gfx.drawFastHLine(x, y, BUTTON_WIDTH);
  gfx.drawFastHLine(x, y + BUTTON_HEIGHT, BUTTON_WIDTH);
  gfx.drawFastVLine(x, y, BUTTON_HEIGHT);
  gfx.drawFastVLine(x + BUTTON_WIDTH, y, BUTTON_HEIGHT);
}

void drawButtons()
{
  gfx.setTextSize(FONT_SIZE_NORMAL);
  for (int i = 0; i <= 11; i++)
  {
    String label = String(i + 1);
    if (i == 9)
    {
      label = String("<");
    }
    if (i == 10)
    {
      label = String("0");
    }
    if (i == 11)
    {
      label = String(">");
    }

    drawButton(i, label);
  }
}

void drawInputField()
{
  int x = 405;
  int y = 20;

  gfx.setColor(TFT_BLACK);
  gfx.drawFastHLine(x, y, 520);
  gfx.drawFastHLine(x, y + 115, 520);
  gfx.drawFastVLine(x, y, 115);
  gfx.drawFastVLine(x + 520, y, 115);
}

void drawInput()
{
  int x = 405;
  int y = 20;

  gfx.setColor(TFT_WHITE);
  gfx.fillRect(x + 1, y + 1, 519, 114);
  gfx.setColor(TFT_BLACK);

  gfx.setEpdMode(epd_mode_t::epd_text);
  gfx.setColor(TFT_BLACK);
  gfx.setTextSize(FONT_SIZE_LARGE);
  gfx.setCursor(x + (520 / 2) - ((gfx.fontWidth() * input.length()) / 2), y + (115 / 2) - (gfx.fontHeight() / 2));
  String display = "";
  for (int i = 0; i < input.length(); i++)
  {
    display += "*";
  }
  gfx.print(display);
  gfx.setEpdMode(epd_mode_t::epd_fast);
}

void setupTime()
{
  timeClient.begin();
}

void send(String state)
{
  if (SETTINGS == nullptr)
  {
    return;
  }

  if (!WiFi.isConnected())
  {
    resetWifi();
  }

  if (client->connect(SETTINGS->gettext("mqtt:ident"), SETTINGS->gettext("mqtt:username"), SETTINGS->gettext("mqtt:password")))
  {
    client->unsubscribe(SETTINGS->gettext("mqtt:state_topic"));
    client->subscribe(SETTINGS->gettext("mqtt:state_topic"));
    client->publish(SETTINGS->gettext("mqtt:command_topic"), state.c_str());
  }
}

void lock()
{
  if (is_disarmed)
  {
    send("ARM_AWAY");
  }
}

void unlock()
{
  send("DISARM");
}

bool isConnecting = false;
void resetWifi()
{
  if (!SETTINGS)
  {
    return;
  }

  if (isConnecting)
  {
    return;
  }
  isConnecting = true;
  drawLockState("connecting");

  WiFi.begin(SETTINGS->gettext("wifi:ssid"), SETTINGS->gettext("wifi:password"));

  if (WiFi.isConnected())
  {
    WiFi.disconnect();
  }

  Serial.print("Connecting to Wi-Fi network");
  for (int cnt_retry = 0; cnt_retry < WIFI_CONNECT_RETRY_MAX && !WiFi.isConnected();
       cnt_retry++)
  {
    delay(500);
    Serial.print(".");
  }
  Serial.println("");
  if (WiFi.isConnected())
  {
    Serial.print("Local IP: ");
    Serial.println(WiFi.localIP());
    isConnecting = false;
  }
  else
  {
    drawLockState("failed to connect");
    isConnecting = false;
    return;
  }

  IPAddress ip;
  if (!ip.fromString(SETTINGS->gettext("mqtt:hostname")))
  {
    drawLockState("failed to parse mqtt ip");
    isConnecting = false;
    return;
  }

  client = new PubSubClient(ip, 1883, mqtt_callback, wifiClient);

  if (client->connect(SETTINGS->gettext("mqtt:ident"), SETTINGS->gettext("mqtt:username"), SETTINGS->gettext("mqtt:password")))
  {
    client->unsubscribe(SETTINGS->gettext("mqtt:state_topic"));
    client->subscribe(SETTINGS->gettext("mqtt:state_topic"));
  }
}

bool initSD()
{
  if (!SD.exists("/settings.yml"))
  {
    Serial.println("settings.yml not found");
    drawLockState("settings.yml not found");
    return false;
  }

  File settingsFile = SD.open("/settings.yml");
  YAMLNode node = YAMLNode::loadStream(settingsFile);
  settingsFile.close();
  SETTINGS = new YAMLNode(node);

  return true;
}

void initTOTP()
{
  uint8_t *base32_key = new uint8_t[20];
  const char *hmac = ((String)SETTINGS->gettext("totp:hmac")).c_str();

  for (int i = 0; i < 20; i++)
  {
    base32_key[i] = (uint8_t)hmac[i];
  }

  TOTP(base32_key, 20, 30);
}

void setup()
{
  M5.begin();

  gfx.init();
  gfx.setEpdMode(epd_mode_t::epd_fast);
  gfx.setRotation(1);
  cls();

  drawButtons();
  drawInputField();
  drawInput();

  if (!initSD())
    return;

  resetWifi();
  sleep(1);
  setupTime();
  initTOTP();

  lock();
}

int index(int x, int y)
{
  for (int index = 0; index <= 11; index++)
  {
    float_t cx = (125 * (index % 3)) + 15;
    float_t cy = (125 * (index / 3)) + 20;

    if (x > cx && x < cx + BUTTON_WIDTH && y > cy && y < cy + BUTTON_HEIGHT)
    {
      return index;
    }
  }

  return -1;
}

void submit(String code)
{
  uint32_t newCode = getCodeFromTimestamp(timeClient.getEpochTime());
  uint32_t input = code.toInt();

  if (newCode == input)
  {
    unlock();
    return;
  }

  lock();
}

void checkForInput()
{
  if (M5.TP.available())
  {
    M5.TP.update();
    if (M5.TP.isFingerUp())
    {
      last_index = -1;
      return;
    }

    tp_finger_t fingerItem = M5.TP.readFinger(0);

    int i = index(fingerItem.x, fingerItem.y);
    if (last_index != i)
    {
      last_index = i;
      if (i == 9)
      {
        input = input.substring(0, input.length() - 1);
      }
      else if (i == 10)
      {
        input += String(0);
      }
      else if (i == 11)
      {
        // submit
        submit(input);
        input = "";
      }
      else
      {
        input += String(i + 1);
      }
    }
  }
}

unsigned long lastBtnPressed = millis();

void checkForButton()
{
  M5.BtnP.read();
  if (M5.BtnP.isPressed() && millis() - lastBtnPressed > 1000)
  {
    lastBtnPressed = millis(); // try and debouce, not really working i guess
    resetWifi();
  }
}

void loop()
{
  checkForButton();
  timeClient.update();
  if (client != nullptr)
  {
    client->loop();
  }
  checkForInput();
  if (oldinput != input)
  {
    oldinput = input;
    drawInput();
  }
}
